@@ -4,7 +4,7 @@ use strict;
 use warnings;
 
 our $VERSION;
-$VERSION = '1.01';
+$VERSION = '1.02';
 
 =head1 NAME
 
@@ -20,14 +20,31 @@ Net::DNS::Resolver::Unbound - Unbound resolver base for Net::DNS
 
 Net::DNS::Resolver::Unbound is designed as an extension to an existing Net::DNS installation.
 
-Net::DNS::Resolver::Unbound replaces the Net::DNS::Resolver::Base implementation.
+Net::DNS::Resolver::Unbound replaces the resolver send() and bgsend() functionality in
+the Net::DNS::Resolver::Base implementation.
+
+As of this writing, the implementation has some significant limitations:
+
+=over 3
+
+=item *
+
+Nameserver selection is entirely at the discretion of Unbound.
+Any user-specified nameservers are ignored.
+
+=item *
+
+It is not possible to send a pre-constructed DNS query packet to a nameserver.
+A best-effort attempt is instead made using the specified (qname,qtype,qclass).
+
+=back
 
 =cut
 
 
-use base qw(Net::DNS::Resolver DynaLoader);
 use Carp;
-use IO::Select;
+use Net::DNS;
+use base qw(Net::DNS::Resolver DynaLoader);
 
 eval { Net::DNS::Resolver::Unbound->bootstrap($VERSION) } || croak $@;
 
@@ -47,7 +64,7 @@ Returns a new Net::DNS::Resolver::Unbound resolver object.
 sub new {
 	my $class = shift;
 	my $self  = $class->SUPER::new();
-	$self->{ub_ctx} = Net::DNS::Resolver::libunbound::ub_ctx_create();
+	$self->{ub_ctx} = Net::DNS::Resolver::Unbound::Context->new();
 	while ( my $attr = shift ) {
 		my $value = shift;
 		my $ref	  = ref($value);
@@ -58,16 +75,8 @@ sub new {
 	return $self;
 }
 
-sub DESTROY {
-	my $self = shift;
-	my $ctx	 = $self->{ub_ctx};
-	Net::DNS::Resolver::libunbound::ub_wait($ctx);
-	Net::DNS::Resolver::libunbound::ub_ctx_delete($ctx);
-	return;
-}
 
-
-=head2 search, query, send, bgsend, bgbusy, bgread, bgcancel
+=head2 search, query, send, bgsend, bgbusy, bgread
 
 See L<Net::DNS::Resolver>.
 
@@ -82,13 +91,12 @@ sub send {
 	my $qtype   = $query->{qtype};
 	my $qclass  = $query->{qclass};
 
-	my $reply = eval {
-		my $ctx	   = $self->{ub_ctx};
-		my $result = Net::DNS::Resolver::libunbound::ub_resolve( $ctx, $qname, $qtype, $qclass );
-		$self->_decode_result( undef, $result );
-	};
+	my $ctx	   = $self->{ub_ctx};
+	my $result = eval { Net::DNS::Resolver::libunbound::ub_resolve( $ctx, $qname, $qtype, $qclass ) };
 	$self->errorstring($@);
-	return $reply;
+	my $packet = $self->_decode_result($result);
+	Net::DNS::Resolver::libunbound::ub_resolve_free($result);
+	return $packet;
 }
 
 sub bgsend {
@@ -100,48 +108,39 @@ sub bgsend {
 	my $qtype   = $query->{qtype};
 	my $qclass  = $query->{qclass};
 
-	my $handle   = {};
-	my $callback = $handle->{anchor} = sub {		# sub{} gets destroyed with $handle
-		$handle->{result} = [@_];
+	my $handle   = [];
+	my $async_id = eval {
+		Net::DNS::Resolver::libunbound::ub_resolve_async( $self->{ub_ctx}, $qname, $qtype, $qclass, $handle );
 	};
-
-	my $ctx = $self->{ub_ctx};
-	$handle->{async_id} =
-			Net::DNS::Resolver::libunbound::ub_resolve_async( $ctx, $qname, $qtype, $qclass, $callback );
+	$self->errorstring($@);
+	unshift( @$handle, $async_id );
 	return $handle;
 }
 
 sub bgbusy {
 	my ( $self, $handle ) = @_;
 	return unless $handle;
-	return if exists( $handle->{result} );
+	my ( undef, @pre ) = @$handle;
+	return if scalar(@pre);
 	my $ctx = $self->{ub_ctx};
-	Net::DNS::Resolver::libunbound::ub_process($ctx)
+	eval { Net::DNS::Resolver::libunbound::ub_process($ctx) }
 			if Net::DNS::Resolver::libunbound::ub_poll($ctx);
-	return !exists( $handle->{result} );
+	$self->errorstring($@);
+	eval { select( undef, undef, undef, 0.200 ) };		# avoid tight loop on bgbusy()
+	my ( undef, @post ) = @$handle;
+	return scalar(@post) ? 0 : 1;
 }
 
 sub bgread {
 	my ( $self, $handle ) = @_;
+
+	eval { Net::DNS::Resolver::libunbound::ub_wait( $self->{ub_ctx} ) } if &bgbusy;
+	$self->errorstring($@);
+
 	return unless $handle;
-
-	Net::DNS::Resolver::libunbound::ub_wait( $self->{ub_ctx} ) if &bgbusy;
-
-	my $result = $handle->{result} || [];
-
-	my $reply = $self->_decode_result(@$result);
-	undef $handle;
-	return $reply;
-}
-
-
-sub bgcancel {
-	my ( $self, $handle ) = @_;
-
-	Net::DNS::Resolver::libunbound::ub_cancel( $self->{ub_ctx}, $handle->{async_id} );
-
-	undef $handle;
-	return;
+	my ( $async_id, $err, $result ) = @$handle;
+	$self->errorstring( Net::DNS::Resolver::libunbound::ub_strerror($err) ) if $err;
+	return $self->_decode_result($result);
 }
 
 
@@ -157,9 +156,9 @@ Get or set Unbound resolver (name,value) context options.
 sub option {
 	my ( $self, $name, @value ) = @_;
 	my $ctx = $self->{ub_ctx};
-	return Net::DNS::Resolver::libunbound::ub_ctx_get_option( $ctx, "$name:" ) unless @value;
-	Net::DNS::Resolver::libunbound::ub_ctx_set_option( $ctx, "$name:", @value );
-	return;
+	return Net::DNS::Resolver::libunbound::ub_ctx_set_option( $ctx, "$name:", @value ) if @value;
+	my $value = Net::DNS::Resolver::libunbound::ub_ctx_get_option( $ctx, $name );
+	return wantarray ? split /\r*\n/, $value : $value;
 }
 
 
@@ -198,20 +197,17 @@ sub replyfrom { return "(local) Unbound resolver" }
 
 
 sub _decode_result {
-	my ( $self, $err, $result ) = @_;
+	my ( $self, $result ) = @_;
 
-	$self->errorstring( Net::DNS::Resolver::libunbound::ub_strerror($err) ) if $err;
-
-	my $answer;
+	my $packet;
 	if ($result) {
-		my $packet = Net::DNS::Resolver::libunbound::ub_result_packet($result);
-		$answer = Net::DNS::Packet->decode( \$packet, $self->debug );
+		my $buffer = Net::DNS::Resolver::libunbound::ub_result_packet($result);
+		$packet = Net::DNS::Packet->decode( \$buffer, $self->debug );
 		$self->errorstring($@);
-		$answer->from( $self->replyfrom );
-		Net::DNS::Resolver::libunbound::ub_resolve_free($result);
 	}
 
-	return $answer;
+	$packet->from( $self->replyfrom ) if $packet;
+	return $packet;
 }
 
 
